MODULE Bessel_I0

! The I0 Bessel functions

IMPLICIT NONE
INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)

PRIVATE
PUBLIC  :: calci0, besi0, besei0

CONTAINS


SUBROUTINE calci0(arg, result, jint)
 
! Code converted using TO_F90 by Alan Miller
! Date: 2003-01-14  Time: 15:25:00
 
!--------------------------------------------------------------------

! This packet computes modified Bessel functions of the first kind
!   and order zero, I0(X) and EXP(-ABS(X))*I0(X), for real
!   arguments X.  It contains two function type subprograms, BESI0
!   and BESEI0, and one subroutine type subprogram, CALCI0.
!   The calling statements for the primary entries are

!                   Y=BESI0(X)
!   and
!                   Y=BESEI0(X)

!   where the entry points correspond to the functions I0(X) and
!   EXP(-ABS(X))*I0(X), respectively.  The routine CALCI0 is
!   intended for internal packet use only, all computations within
!   the packet being concentrated in this routine.  The function
!   subprograms invoke CALCI0 with the statement
!          CALL CALCI0(ARG,RESULT,JINT)
!   where the parameter usage is as follows

!      Function                     Parameters for CALCI0
!       Call              ARG                  RESULT          JINT

!     BESI0(ARG)    ABS(ARG) .LE. XMAX        I0(ARG)           1
!     BESEI0(ARG)    any real ARG        EXP(-ABS(ARG))*I0(ARG) 2

!   The main computation evaluates slightly modified forms of
!   minimax approximations generated by Blair and Edwards, Chalk
!   River (Atomic Energy of Canada Limited) Report AECL-4928,
!   October, 1974.  This transportable program is patterned after
!   the machine-dependent FUNPACK packet NATSI0, but cannot match
!   that version for efficiency or accuracy.  This version uses
!   rational functions that theoretically approximate I-SUB-0(X)
!   to at least 18 significant decimal digits.  The accuracy
!   achieved depends on the arithmetic system, the compiler, the
!   intrinsic functions, and proper selection of the machine-
!   dependent constants.

!*******************************************************************

! Explanation of machine-dependent constants.  Let

!   beta   = Radix for the floating-point system
!   maxexp = Smallest power of beta that overflows

! Then the following machine-dependent constants must be declared
!   in DATA statements.  IEEE values are provided as a default.

!   XSMALL = Positive argument such that 1.0 - X = 1.0 to
!            machine precision for all ABS(X) .LE. XSMALL.
!   XINF =   Largest positive machine number; approximately beta**maxexp
!   XMAX =   Largest argument acceptable to BESI0;  Solution to
!            equation:
!               W(X) * (1+1/(8*X)+9/(128*X**2) = beta**maxexp
!            where  W(X) = EXP(X)/SQRT(2*PI*X)


!     Approximate values for some important machines are:

!                          beta       maxexp       XSMALL

! CRAY-1        (S.P.)       2         8191       3.55E-15
! Cyber 180/855
!   under NOS   (S.P.)       2         1070       3.55E-15
! IEEE (IBM/XT,
!   SUN, etc.)  (S.P.)       2          128       2.98E-8
! IEEE (IBM/XT,
!   SUN, etc.)  (D.P.)       2         1024       5.55D-17
! IBM 3033      (D.P.)      16           63       6.95D-18
! VAX           (S.P.)       2          127       2.98E-8
! VAX D-Format  (D.P.)       2          127       6.95D-18
! VAX G-Format  (D.P.)       2         1023       5.55D-17


!                               XINF          XMAX

! CRAY-1        (S.P.)       5.45E+2465     5682.810
! Cyber 180/855
!   under NOS   (S.P.)       1.26E+322       745.893
! IEEE (IBM/XT,
!   SUN, etc.)  (S.P.)       3.40E+38         91.900
! IEEE (IBM/XT,
!   SUN, etc.)  (D.P.)       1.79D+308       713.986
! IBM 3033      (D.P.)       7.23D+75        178.182
! VAX           (S.P.)       1.70D+38         91.203
! VAX D-Format  (D.P.)       1.70D+38         91.203
! VAX G-Format  (D.P.)       8.98D+307       713.293

!*******************************************************************

! Error returns

!  The program returns XINF for BESI0 for ABS(ARG) .GT. XMAX.


!  Intrinsic functions required are:

!     ABS, SQRT, EXP


!  Authors: W. J. Cody and L. Stoltz
!           Mathematics and Computer Science Division
!           Argonne National Laboratory
!           Argonne, IL 60439

!  Latest modification: March 12, 1992

!--------------------------------------------------------------------

IMPLICIT NONE
INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)

REAL (dp), INTENT(IN)   :: arg
REAL (dp), INTENT(OUT)  :: result
INTEGER, INTENT(IN)     :: jint

! Local variables

INTEGER    :: i
REAL (dp)  :: a, b, sump, sumq, x, xx
!--------------------------------------------------------------------
!  Mathematical constants
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: one = 1.0_dp, one5 = 15.0_dp, forty = 40.0_dp,  &
                         exp40 = 2.353852668370199854D17, two25 = 225.0_dp, &
                         rec15 = 6.6666666666666666666D-2
!--------------------------------------------------------------------
!  Machine-dependent constants
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: XSMALL = 5.55D-17, XINF = 1.79D308, XMAX = 713.986D0
!--------------------------------------------------------------------
!  Coefficients for XSMALL .LE. ABS(ARG) .LT. 15.0
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: p(15) =  &
        (/ -5.2487866627945699800D-18, -1.5982226675653184646D-14,  &
           -2.6843448573468483278D-11, -3.0517226450451067446D-08,  &
           -2.5172644670688975051D-05, -1.5453977791786851041D-02,  &
           -7.0935347449210549190D+00, -2.4125195876041896775D+03,  &
           -5.9545626019847898221D+05, -1.0313066708737980747D+08,  &
           -1.1912746104985237192D+10, -8.4925101247114157499D+11,  &
           -3.2940087627407749166D+13, -5.5050369673018427753D+14,  &
           -2.2335582639474375249D+15 /)
REAL (dp), PARAMETER  :: Q(5) =  &
        (/ -3.7277560179962773046D+03, 6.5158506418655165707D+06,  &
           -6.5626560740833869295D+09, 3.7604188704092954661D+12,  &
           -9.7087946179594019126D+14 /)
!--------------------------------------------------------------------
!  Coefficients for 15.0 .LE. ABS(ARG)
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: PP(8) =  &
        (/ -3.9843750000000000000D-01, 2.9205384596336793945D+00,  &
           -2.4708469169133954315D+00, 4.7914889422856814203D-01,  &
           -3.7384991926068969150D-03,-2.6801520353328635310D-03,  &
            9.9168777670983678974D-05,-2.1877128189032726730D-06 /)
REAL (dp), PARAMETER  :: QQ(7) =  &
        (/ -3.1446690275135491500D+01, 8.5539563258012929600D+01,  &
           -6.0228002066743340583D+01, 1.3982595353892851542D+01,  &
           -1.1151759188741312645D+00, 3.2547697594819615062D-02,  &
           -5.5194330231005480228D-04 /)
!--------------------------------------------------------------------

x = ABS(arg)
IF (x < xsmall) THEN
  result = one
ELSE IF (x < one5) THEN
!--------------------------------------------------------------------
!  XSMALL .LE.  ABS(ARG)  .LT. 15.0
!--------------------------------------------------------------------
  xx = x * x
  sump = p(1)
  DO  i = 2, 15
    sump = sump * xx + p(i)
  END DO
  xx = xx - two25
  sumq = ((((xx+q(1))*xx+q(2))*xx+q(3))*xx+q(4)) * xx + q(5)
  result = sump / sumq
  IF (jint == 2) result = result * EXP(-x)
ELSE IF (x >= one5) THEN
  IF ((jint == 1).AND.(x > xmax)) THEN
    result = xinf
  ELSE
!--------------------------------------------------------------------
!  15.0  .LE.  ABS(ARG)
!--------------------------------------------------------------------
    xx = one / x - rec15
    sump = ((((((pp(1)*xx+pp(2))*xx+pp(3))*xx+pp(4))*xx+pp(5))*xx+  &
        pp(6))*xx+pp(7)) * xx + pp(8)
    sumq = ((((((xx+qq(1))*xx+qq(2))*xx+qq(3))*xx+qq(4))*xx+  &
        qq(5))*xx+qq(6)) * xx + qq(7)
    result = sump / sumq
    IF (jint == 2) THEN
      result = (result-pp(1)) / SQRT(x)
    ELSE
!--------------------------------------------------------------------
!  Calculation reformulated to avoid premature overflow
!--------------------------------------------------------------------
      IF (x <= (xmax-one5)) THEN
        a = EXP(x)
        b = one
      ELSE
        a = EXP(x-forty)
        b = exp40
      END IF
      result = ((result*a-pp(1)*a)/SQRT(x)) * b
    END IF
  END IF
END IF
!--------------------------------------------------------------------
!  Return for ABS(ARG) .LT. XSMALL
!--------------------------------------------------------------------
RETURN
!----------- Last line of CALCI0 -----------
END SUBROUTINE calci0



FUNCTION BESI0(X) RESULT(fn_val)
!--------------------------------------------------------------------

! This long precision subprogram computes approximate values for
!   modified Bessel functions of the first kind of order zero for
!   arguments ABS(ARG) .LE. XMAX  (see comments heading CALCI0).

!--------------------------------------------------------------------

REAL (dp), INTENT(IN)  :: x
REAL (dp)              :: fn_val

! Local variable
INTEGER  :: jint

!--------------------------------------------------------------------
jint = 1
CALL calci0(x, fn_val, jint)
RETURN
!---------- Last line of BESI0 ----------
END FUNCTION BESI0



FUNCTION BESEI0(X) RESULT(fn_val)
!--------------------------------------------------------------------

! This function program computes approximate values for the
!   modified Bessel function of the first kind of order zero
!   multiplied by EXP(-ABS(X)), where EXP is the
!   exponential function, ABS is the absolute value, and X
!   is any argument.

!--------------------------------------------------------------------

REAL (dp), INTENT(IN)  :: x
REAL (dp)              :: fn_val

! Local variable
INTEGER  :: jint

!--------------------------------------------------------------------
jint = 2
CALL calci0(x, fn_val, jint)
RETURN
!---------- Last line of BESEI0 ----------
END FUNCTION BESEI0

END MODULE Bessel_I0
