MODULE Bessel_K0

! The K0 Bessel functions

IMPLICIT NONE
INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)

PRIVATE
PUBLIC  :: calck0, besk0, besek0

CONTAINS



SUBROUTINE calck0(arg, result, jint)
 
! Code converted using TO_F90 by Alan Miller
! Date: 2003-01-14  Time: 15:25:00
 
!--------------------------------------------------------------------

! This packet computes modified Bessel functions of the second kind
!   and order zero, K0(X) and EXP(X)*K0(X), for real arguments X.
!   It contains two function type subprograms, BESK0 and BESEK0,
!   and one subroutine type subprogram, CALCK0.
!   The calling statements for the primary entries are

!                   Y=BESK0(X)
!   and
!                   Y=BESEK0(X)

!   where the entry points correspond to the functions K0(X) and
!   EXP(X)*K0(X), respectively.  The routine CALCK0 is
!   intended for internal packet use only, all computations within
!   the packet being concentrated in this routine.  The function
!   subprograms invoke CALCK0 with the statement
!          CALL CALCK0(ARG,RESULT,JINT)
!   where the parameter usage is as follows

!      Function                     Parameters for CALCK0
!       Call              ARG                  RESULT          JINT

!     BESK0(ARG)   0 .LT. ARG .LE. XMAX       K0(ARG)           1
!     BESEK0(ARG)     0 .LT. ARG           EXP(ARG)*K0(ARG)     2

!   The main computation evaluates slightly modified forms of near
!   minimax rational approximations generated by Russon and Blair,
!   Chalk River (Atomic Energy of Canada Limited) Report AECL-3461,
!   1969.  This transportable program is patterned after the
!   machine-dependent FUNPACK packet NATSK0, but cannot match that
!   version for efficiency or accuracy.  This version uses rational
!   functions that theoretically approximate K-SUB-0(X) to at
!   least 18 significant decimal digits.  The accuracy achieved
!   depends on the arithmetic system, the compiler, the intrinsic
!   functions, and proper selection of the machine-dependent constants.

!*******************************************************************

! Explanation of machine-dependent constants.  Let

!   beta   = Radix for the floating-point system
!   minexp = Smallest representable power of beta
!   maxexp = Smallest power of beta that overflows

! Then the following machine-dependent constants must be declared
!   in DATA statements.  IEEE values are provided as a default.

!   XSMALL = Argument below which BESK0 and BESEK0 may
!            each be represented by a constant and a log.
!            largest X such that  1.0 + X = 1.0  to machine precision.
!   XINF   = Largest positive machine number; approximately beta**maxexp
!   XMAX   = Largest argument acceptable to BESK0;  Solution to
!            equation:
!               W(X) * (1-1/8X+9/128X**2) = beta**minexp
!            where  W(X) = EXP(-X)*SQRT(PI/2X)


!     Approximate values for some important machines are:


!                           beta       minexp       maxexp

!  CRAY-1        (S.P.)       2        -8193         8191
!  Cyber 180/185
!    under NOS   (S.P.)       2         -975         1070
!  IEEE (IBM/XT,
!    SUN, etc.)  (S.P.)       2         -126          128
!  IEEE (IBM/XT,
!    SUN, etc.)  (D.P.)       2        -1022         1024
!  IBM 3033      (D.P.)      16          -65           63
!  VAX D-Format  (D.P.)       2         -128          127
!  VAX G-Format  (D.P.)       2        -1024         1023


!                          XSMALL       XINF         XMAX

! CRAY-1        (S.P.)    3.55E-15   5.45E+2465    5674.858
! Cyber 180/855
!   under NOS   (S.P.)    1.77E-15   1.26E+322      672.788
! IEEE (IBM/XT,
!   SUN, etc.)  (S.P.)    5.95E-8    3.40E+38        85.337
! IEEE (IBM/XT,
!   SUN, etc.)  (D.P.)    1.11D-16   1.79D+308      705.342
! IBM 3033      (D.P.)    1.11D-16   7.23D+75       177.852
! VAX D-Format  (D.P.)    6.95D-18   1.70D+38        86.715
! VAX G-Format  (D.P.)    5.55D-17   8.98D+307      706.728

!*******************************************************************

! Error returns

!  The program returns the value XINF for ARG .LE. 0.0, and the
!  BESK0 entry returns the value 0.0 for ARG .GT. XMAX.


!  Intrinsic functions required are:

!     EXP, LOG, SQRT

!  Latest modification: March 14, 1992

!  Authors: W. J. Cody and Laura Stoltz
!           Mathematics and Computer Science Division
!           Argonne National Laboratory
!           Argonne, IL 60439, USA

!--------------------------------------------------------------------

REAL (dp), INTENT(IN)   :: arg
REAL (dp), INTENT(OUT)  :: result
INTEGER, INTENT(IN)     :: jint

! Local variables
INTEGER    :: i
REAL (dp)  :: sumf, sumg, sump, sumq, temp, x, xx
!--------------------------------------------------------------------
!  Mathematical constants
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: one = 1.0_dp, zero = 0.0_dp
!--------------------------------------------------------------------
!  Machine-dependent constants
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: XSMALL = 1.11D-16, XINF = 1.79D+308, XMAX = 705.342D0
!--------------------------------------------------------------------

!     Coefficients for XSMALL .LE.  ARG  .LE. 1.0

!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: P(6) =   &
           (/  5.8599221412826100000D-04, 1.3166052564989571850D-01,  &
               1.1999463724910714109D+01, 4.6850901201934832188D+02,  &
               5.9169059852270512312D+03, 2.4708152720399552679D+03 /)
REAL (dp), PARAMETER  :: Q(2) =   &
           (/ -2.4994418972832303646D+02, 2.1312714303849120380D+04 /)
REAL (dp), PARAMETER  :: F(4) =  &
           (/ -1.6414452837299064100D+00, -2.9601657892958843866D+02,  &
              -1.7733784684952985886D+04, -4.0320340761145482298D+05 /)
REAL (dp), PARAMETER  :: G(3) =  &
           (/ -2.5064972445877992730D+02, 2.9865713163054025489D+04,  &
              -1.6128136304458193998D+06 /)
!--------------------------------------------------------------------

!     Coefficients for  1.0 .LT. ARG

!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: PP(10) =  &
           (/ 1.1394980557384778174D+02, 3.6832589957340267940D+03,  &
              3.1075408980684392399D+04, 1.0577068948034021957D+05,  &
              1.7398867902565686251D+05, 1.5097646353289914539D+05,  &
              7.1557062783764037541D+04, 1.8321525870183537725D+04,  &
              2.3444738764199315021D+03, 1.1600249425076035558D+02 /)
REAL (dp), PARAMETER  :: QQ(10) =  &
           (/ 2.0013443064949242491D+02, 4.4329628889746408858D+03,  &
              3.1474655750295278825D+04, 9.7418829762268075784D+04,  &
              1.5144644673520157801D+05, 1.2689839587977598727D+05,  &
              5.8824616785857027752D+04, 1.4847228371802360957D+04,  &
              1.8821890840982713696D+03, 9.2556599177304839811D+01 /)
!--------------------------------------------------------------------

x = arg
IF (x > zero) THEN
  IF (x <= one) THEN
!--------------------------------------------------------------------
!     0.0 .LT.  ARG  .LE. 1.0
!--------------------------------------------------------------------
    temp = LOG(x)
    IF (x < xsmall) THEN
!--------------------------------------------------------------------
!     Return for small ARG
!--------------------------------------------------------------------
      result = p(6) / q(2) - temp
    ELSE
      xx = x * x
      sump = ((((p(1)*xx+p(2))*xx+p(3))*xx+p(4))*xx+p(5)) * xx + p(6)
      sumq = (xx+q(1)) * xx + q(2)
      sumf = ((f(1)*xx+f(2))*xx+f(3)) * xx + f(4)
      sumg = ((xx+g(1))*xx+g(2)) * xx + g(3)
      result = sump / sumq - xx * sumf * temp / sumg - temp
      IF (jint == 2) result = result * EXP(x)
    END IF
  ELSE IF ((jint == 1) .AND. (x > xmax)) THEN
!--------------------------------------------------------------------
!     Error return for ARG .GT. XMAX
!--------------------------------------------------------------------
    result = zero
  ELSE
!--------------------------------------------------------------------
!     1.0 .LT. ARG
!--------------------------------------------------------------------
    xx = one / x
    sump = pp(1)
    DO  i = 2, 10
      sump = sump * xx + pp(i)
    END DO
    sumq = xx
    DO  i = 1, 9
      sumq = (sumq+qq(i)) * xx
    END DO
    sumq = sumq + qq(10)
    result = sump / sumq / SQRT(x)
    IF (jint == 1) result = result * EXP(-x)
  END IF
ELSE
!--------------------------------------------------------------------
!     Error return for ARG .LE. 0.0
!--------------------------------------------------------------------
  result = xinf
END IF
!--------------------------------------------------------------------
!     Update error counts, etc.
!--------------------------------------------------------------------
RETURN
!---------- Last line of CALCK0 ----------
END SUBROUTINE calck0



FUNCTION BESK0(X) RESULT(fn_val)
!--------------------------------------------------------------------

! This function program computes approximate values for the
!   modified Bessel function of the second kind of order zero
!   for arguments 0.0 .LT. ARG .LE. XMAX (see comments heading
!   CALCK0).

!  Authors: W. J. Cody and Laura Stoltz

!  Latest Modification: March 14, 1992

!--------------------------------------------------------------------
REAL (dp), INTENT(IN)  :: x
REAL (dp)              :: fn_val

! Local variable
INTEGER  :: jint
!--------------------------------------------------------------------
jint = 1
CALL calck0(x, fn_val, jint)
RETURN
!---------- Last line of BESK0 ----------
END FUNCTION BESK0



FUNCTION BESEK0(X) RESULT(fn_val)
!--------------------------------------------------------------------

! This function program computes approximate values for the
!   modified Bessel function of the second kind of order zero
!   multiplied by the Exponential function, for arguments
!   0.0 .LT. ARG.

!  Authors: W. J. Cody and Laura Stoltz

!  Latest Modification: March 14, 1992

!--------------------------------------------------------------------

REAL (dp), INTENT(IN)  :: x
REAL (dp)              :: fn_val

! Local variable
INTEGER  :: jint
!--------------------------------------------------------------------
jint = 2
CALL calck0(x, fn_val, jint)
RETURN
!---------- Last line of BESEK0 ----------
END FUNCTION BESEK0

END MODULE Bessel_K0
