MODULE Bessel_K1

! The K1 Bessel functions

IMPLICIT NONE
INTEGER, PARAMETER  :: dp = SELECTED_REAL_KIND(12, 60)

PRIVATE
PUBLIC  :: calck1, besk1, besek1

CONTAINS


SUBROUTINE calck1(arg, result, jint)
 
! Code converted using TO_F90 by Alan Miller
! Date: 2003-01-14  Time: 15:25:00
 
!--------------------------------------------------------------------

! This packet computes modified Bessel functions of the second kind and
!   order one,  K1(X)  and  EXP(X)*K1(X), for real arguments X.
!   It contains two function type subprograms, BESK1  and  BESEK1,
!   and one subroutine type subprogram, CALCK1.  The calling
!   statements for the primary entries are

!                   Y=BESK1(X)
!   and
!                   Y=BESEK1(X)

!   where the entry points correspond to the functions K1(X) and
!   EXP(X)*K1(X), respectively.  The routine CALCK1 is intended
!   for internal packet use only, all computations within the
!   packet being concentrated in this routine.  The function
!   subprograms invoke CALCK1 with the statement
!          CALL CALCK1(ARG,RESULT,JINT)
!   where the parameter usage is as follows

!      Function                      Parameters for CALCK1
!        Call             ARG                  RESULT        JINT

!     BESK1(ARG)   XLEAST < ARG < XMAX      K1(ARG)           1
!     BESEK1(ARG)      XLEAST < ARG       EXP(ARG)*K1(ARG)    2

!   The main computation evaluates slightly modified forms of near
!   minimax rational approximations generated by Russon and Blair,
!   Chalk River (Atomic Energy of Canada Limited) Report AECL-3461,
!   1969.  This transportable program is patterned after the
!   machine-dependent FUNPACK packet NATSK1, but cannot match that
!   version for efficiency or accuracy.  This version uses rational
!   functions that theoretically approximate K-SUB-1(X) to at
!   least 18 significant decimal digits.  The accuracy achieved
!   depends on the arithmetic system, the compiler, the intrinsic
!   functions, and proper selection of the machine-dependent constants.

!*******************************************************************

! Explanation of machine-dependent constants.  Let

!   beta   = Radix for the floating-point system
!   minexp = Smallest representable power of beta
!   maxexp = Smallest power of beta that overflows

! Then the following machine-dependent constants must be declared
!   in DATA statements.  IEEE values are provided as a default.

!   XLEAST = Smallest acceptable argument, i.e., smallest machine
!            number X such that 1/X is machine representable.
!   XSMALL = Argument below which BESK1(X) and BESEK1(X) may
!            each be represented by 1/X.  A safe value is the
!            largest X such that  1.0 + X = 1.0  to machine precision.
!   XINF   = Largest positive machine number; approximately
!            beta**maxexp
!   XMAX   = Largest argument acceptable to BESK1;  Solution to
!            equation:
!               W(X) * (1+3/8X-15/128X**2) = beta**minexp
!            where  W(X) = EXP(-X)*SQRT(PI/2X)


!     Approximate values for some important machines are:

!                           beta       minexp       maxexp

!  CRAY-1        (S.P.)       2        -8193         8191
!  Cyber 180/185
!    under NOS   (S.P.)       2         -975         1070
!  IEEE (IBM/XT,
!    SUN, etc.)  (S.P.)       2         -126          128
!  IEEE (IBM/XT,
!    SUN, etc.)  (D.P.)       2        -1022         1024
!  IBM 3033      (D.P.)      16          -65           63
!  VAX D-Format  (D.P.)       2         -128          127
!  VAX G-Format  (D.P.)       2        -1024         1023


!                         XLEAST     XSMALL      XINF       XMAX

! CRAY-1                1.84E-2466  3.55E-15  5.45E+2465  5674.858
! Cyber 180/855
!   under NOS   (S.P.)  3.14E-294   1.77E-15  1.26E+322    672.789
! IEEE (IBM/XT,
!   SUN, etc.)  (S.P.)  1.18E-38    5.95E-8   3.40E+38      85.343
! IEEE (IBM/XT,
!   SUN, etc.)  (D.P.)  2.23D-308   1.11D-16  1.79D+308    705.343
! IBM 3033      (D.P.)  1.39D-76    1.11D-16  7.23D+75     177.855
! VAX D-Format  (D.P.)  5.88D-39    6.95D-18  1.70D+38      86.721
! VAX G-Format  (D.P.)  1.12D-308   5.55D-17  8.98D+307    706.728

!*******************************************************************

! Error returns

!  The program returns the value XINF for ARG .LE. 0.0 and the
!   BESK1 entry returns the value 0.0 for ARG > XMAX.


!  Intrinsic functions required are:

!     LOG, SQRT, EXP


!  Authors: W. J. Cody and Laura Stoltz
!           Mathematics and Computer Science Division
!           Argonne National Laboratory
!           Argonne, IL 60439, USA

!  Latest modification: March 14, 1992

!--------------------------------------------------------------------

REAL (dp), INTENT(IN)   :: arg
REAL (dp), INTENT(OUT)  :: result
INTEGER, INTENT(IN)     :: jint

INTEGER    :: i
REAL (dp)  :: sumf, sumg, sump, sumq, x, xx
!--------------------------------------------------------------------
!  Mathematical constants
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: one = 1.0_dp, zero = 0.0_dp
!--------------------------------------------------------------------
!  Machine-dependent constants
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: XLEAST = 2.23D-308, XSMALL = 1.11D-16,   &
                         XINF = 1.79D+308, XMAX = 705.343_dp
!--------------------------------------------------------------------
!  Coefficients for  XLEAST <=  ARG  <= 1.0
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: p(5) =   &
           (/ 4.8127070456878442310D-1, 9.9991373567429309922D+1,  &
              7.1885382604084798576D+3, 1.7733324035147015630D+5,  &
              7.1938920065420586101D+5 /)
REAL (dp), PARAMETER  :: q(3) =  &
           (/ -2.8143915754538725829D+2, 3.7264298672067697862D+4,  &
              -2.2149374878243304548D+6 /)
REAL (dp), PARAMETER  :: f(5) =  &
           (/ -2.2795590826955002390D-1, -5.3103913335180275253D+1,  &
              -4.5051623763436087023D+3, -1.4758069205414222471D+5,  &
              -1.3531161492785421328D+6 /)
REAL (dp), PARAMETER  :: g(3) =  &
           (/ -3.0507151578787595807D+2, 4.3117653211351080007D+4,  &
              -2.7062322985570842656D+6 /)
!--------------------------------------------------------------------
!  Coefficients for  1.0 <  ARG
!--------------------------------------------------------------------
REAL (dp), PARAMETER  :: pp(11) =  &
           (/ 6.4257745859173138767D-2, 7.5584584631176030810D+0,  &
              1.3182609918569941308D+2, 8.1094256146537402173D+2,  &
              2.3123742209168871550D+3, 3.4540675585544584407D+3,  &
              2.8590657697910288226D+3, 1.3319486433183221990D+3,  &
              3.4122953486801312910D+2, 4.4137176114230414036D+1,  &
              2.2196792496874548962D+0 /)
REAL (dp), PARAMETER  :: qq(9) =  &
           (/ 3.6001069306861518855D+1, 3.3031020088765390854D+2,  &
              1.2082692316002348638D+3, 2.1181000487171943810D+3,  &
              1.9448440788918006154D+3, 9.6929165726802648634D+2,  &
              2.5951223655579051357D+2, 3.4552228452758912848D+1,  &
              1.7710478032601086579D+0 /)
!--------------------------------------------------------------------

x = arg
IF (x < xleast) THEN
!--------------------------------------------------------------------
!  Error return for  ARG  < XLEAST
!--------------------------------------------------------------------
  result = xinf
ELSE IF (x <= one) THEN
!--------------------------------------------------------------------
!  XLEAST <=  ARG  <= 1.0
!--------------------------------------------------------------------
  IF (x < xsmall) THEN
!--------------------------------------------------------------------
!  Return for small ARG
!--------------------------------------------------------------------
    result = one / x
  ELSE
    xx = x * x
    sump = ((((p(1)*xx+p(2))*xx+p(3))*xx+p(4))*xx+p(5)) * xx + q(3)
    sumq = ((xx+q(1))*xx+q(2)) * xx + q(3)
    sumf = (((f(1)*xx+f(2))*xx+f(3))*xx+f(4)) * xx + f(5)
    sumg = ((xx+g(1))*xx+g(2)) * xx + g(3)
    result = (xx*LOG(x)*sumf/sumg+sump/sumq) / x
    IF (jint == 2) result = result * EXP(x)
  END IF
ELSE IF ((jint == 1).AND.(x > xmax)) THEN
!--------------------------------------------------------------------
!  Error return for  ARG  > XMAX
!--------------------------------------------------------------------
  result = zero
ELSE
!--------------------------------------------------------------------
!  1.0 <  ARG
!--------------------------------------------------------------------
  xx = one / x
  sump = pp(1)
  DO  i = 2, 11
    sump = sump * xx + pp(i)
  END DO
  sumq = xx
  DO  i = 1, 8
    sumq = (sumq+qq(i)) * xx
  END DO
  sumq = sumq + qq(9)
  result = sump / sumq / SQRT(x)
  IF (jint == 1) result = result * EXP(-x)
END IF
RETURN
!---------- Last line of CALCK1 ----------
END SUBROUTINE calck1



FUNCTION BESK1(X) RESULT(fn_val)
!--------------------------------------------------------------------

! This function program computes approximate values for the
!   modified Bessel function of the second kind of order one
!   for arguments  XLEAST <= ARG <= XMAX.

!--------------------------------------------------------------------
REAL (dp), INTENT(IN)  :: x
REAL (dp)              :: fn_val

! Local variable
INTEGER  :: jint
!--------------------------------------------------------------------
jint = 1
CALL calck1(x, fn_val, jint)
RETURN
!---------- Last line of BESK1 ----------
END FUNCTION BESK1



FUNCTION BESEK1(X) RESULT(fn_val)
!--------------------------------------------------------------------

! This function program computes approximate values for the modified
!   Bessel function of the second kind of order one multiplied by the
!   exponential function, for arguments  XLEAST <= ARG <= XMAX.

!--------------------------------------------------------------------
REAL (dp), INTENT(IN)  :: x
REAL (dp)              :: fn_val

! Local variable
INTEGER  :: jint
!--------------------------------------------------------------------
jint = 2
CALL calck1(x, fn_val, jint)
RETURN
!---------- Last line of BESEK1 ----------
END FUNCTION BESEK1

END MODULE Bessel_K1
